#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include<iostream>
// dune includes
#include<dune/common/parallel/mpihelper.hh>
#include<dune/common/parametertreeparser.hh>
#include<dune/common/timer.hh>
//#if HAVE_UG
//#include<dune/grid/uggrid.hh>
//#endif
#if HAVE_DUNE_ALUGRID
#include<dune/alugrid/grid.hh>
#include<dune/alugrid/dgf.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#endif
// pdelab includes
#include<dune/pdelab/finiteelementmap/pkfem.hh>

#include "2p_driver.hh"

//===============================================================
// Main program with grid setup
//===============================================================




int main(int argc, char** argv)
{
  #if !HAVE_SUPERLU
    std::cout << "Error: These examples work only if SuperLU is available." << std::endl;
    exit(1);
  #endif

  #if HAVE_UG
  try{
    // Maybe initialize Mpi
    Dune::MPIHelper&
      helper = Dune::MPIHelper::instance(argc, argv);
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout << "Parallel code run on "
                << helper.size() << " process(es)" << std::endl;

    // open ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("ss_calcite.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);

    typedef Dune::ALUGrid<2,2,Dune::simplex,Dune::conforming> Grid;
    //typedef Dune::UGGrid<2> Grid;
    typedef Grid::ctype DF;

    std::vector<int> boundary_index_map;
    std::vector<int> element_index_map;

    std::string filename = ptree.get("domain.filename",
                                     "turbtube2d.msh");
    Dune::GridFactory<Grid> factory;
    if (helper.rank()==0)
      Dune::GmshReader<Grid>::read(factory,filename,boundary_index_map,element_index_map,true,false);
    std::shared_ptr<Grid> gridp(factory.createGrid());

    typedef Grid::LeafGridView GV;
    GV gv = gridp->leafGridView();

    std::cout << "OK" << std::endl;

    using ES = Dune::PDELab::NonOverlappingEntitySet<GV>;
    ES es(gv);

    const int k = 2;
    typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k> V_FEM;
    typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,k-1> P_FEM;
    const int l = 2;
    typedef Dune::PDELab::PkLocalFiniteElementMap<ES,DF,double,l> PHI_FEM;

    V_FEM vFem(es);
    P_FEM pFem(es);
    PHI_FEM phiFem(es);

    driver_fs(*gridp, es,vFem,pFem,phiFem, ptree, helper);


  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
    return 1;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
    return 1;
  }
  #endif
}
